/* -*- Mode: C; indent-tabs-mode: t; tab-width: 4 -*-
// ---------------------------------------------------------------------------
// SquirrelJME
//     Copyright (C) Stephanie Gawroriski <xer@multiphasicapps.net>
// ---------------------------------------------------------------------------
// SquirrelJME is under the Mozilla Public License Version 2.0.
// See license.mkd for licensing and copyright information.
// -------------------------------------------------------------------------*/

#include <string.h>

#include "sjme/charSeq.h"
#include "sjme/debug.h"
#include "sjme/util.h"

static sjme_errorCode sjme_charSeq_basicUtf_charAt(
	sjme_attrInNotNull const sjme_charSeq* inSeq,
	sjme_attrInPositive sjme_jint inIndex,
	sjme_attrOutNotNull sjme_jchar* outChar)
{
	sjme_jint c;
	
	if (inSeq == NULL || outChar == NULL)
		return SJME_ERROR_NULL_ARGUMENTS;

	/* Obtain UTF character at position. */
	c = sjme_string_charAt(inSeq->context, inIndex);
	if (c < 0)
		return SJME_ERROR_ILLEGAL_STATE;
	
	/* Success! */
	*outChar = c;
	return SJME_ERROR_NONE;
}

static sjme_errorCode sjme_charSeq_basicUtf_length(
	sjme_attrInNotNull const sjme_charSeq* inSeq,
	sjme_attrOutNotNull sjme_jint* outLen)
{
	sjme_jint len;
	
	if (inSeq == NULL || outLen == NULL)
		return SJME_ERROR_NULL_ARGUMENTS;
	
	/* Determine string length. */
	len = sjme_string_length(inSeq->context);
	if (len < 0)
		return SJME_ERROR_ILLEGAL_STATE;
	
	/* Success! */
	*outLen = len;
	return SJME_ERROR_NONE;
}

static const sjme_charSeq_functions sjme_charSeq_basicUtfFunctions =
{
	.charAt = sjme_charSeq_basicUtf_charAt,
	.length = sjme_charSeq_basicUtf_length,
};

sjme_errorCode sjme_charSeq_charAt(
	sjme_attrInNotNull const sjme_charSeq* inSeq,
	sjme_attrInPositive sjme_jint inIndex,
	sjme_attrOutNotNull sjme_jchar* outChar)
{
	if (inSeq == NULL || outChar == NULL)
		return SJME_ERROR_NULL_ARGUMENTS;
	
	if (inIndex < 0)
		return SJME_ERROR_INDEX_OUT_OF_BOUNDS;
	
	if (inSeq->impl->charAt == NULL)
		return sjme_error_notImplemented(0);
	
	/* Forward. */
	return inSeq->impl->charAt(inSeq, inIndex, outChar);
}

sjme_errorCode sjme_charSeq_deleteStatic(
	sjme_attrInNotNull sjme_charSeq* inOutSeq)
{
	sjme_errorCode error;
	
	if (inOutSeq == NULL)
		return SJME_ERROR_NULL_ARGUMENTS;
	
	/* Already deleted? */
	if (inOutSeq->impl == NULL)
		return SJME_ERROR_ILLEGAL_STATE;
	
	/* If there is a deletion function, then call it. */
	if (inOutSeq->impl->delete != NULL)
		if (sjme_error_is(error = inOutSeq->impl->delete(
			inOutSeq)))
			return sjme_error_default(error);
	
	/* Clear out. */
	memset(inOutSeq, 0, sizeof(*inOutSeq));
	return SJME_ERROR_NONE;
}

sjme_errorCode sjme_charSeq_equalsCharSeq(
	sjme_attrInNotNull const sjme_charSeq* inSeq,
	sjme_attrOutNotNull sjme_jboolean* outResult,
	sjme_attrInNotNull const sjme_charSeq* equalsSeq)
{
	sjme_errorCode error;
	sjme_jint aLen, bLen;
	sjme_jint at;
	sjme_jchar a, b;
	
	if (inSeq == NULL || outResult == NULL || equalsSeq == NULL)
		return SJME_ERROR_NULL_ARGUMENTS;
	
	/* Same sequence? */
	if (inSeq == equalsSeq)
	{
		*outResult = SJME_JNI_TRUE;
		return SJME_ERROR_NONE;
	}
	
	/* Get length of both first. */
	aLen = -1;
	bLen = -1;
	if (sjme_error_is(error = sjme_charSeq_length(inSeq,
		&aLen)) || aLen < 0)
		return sjme_error_default(error);
	if (sjme_error_is(error = sjme_charSeq_length(equalsSeq,
		&bLen)) || bLen < 0)
		return sjme_error_default(error);
	
	/* Cannot possibly be equal? */
	if (aLen != bLen)
	{
		*outResult = SJME_JNI_FALSE;
		return SJME_ERROR_NONE;
	}
	
	/* Compare each character. */
	for (at = 0; at < aLen; at++)
	{
		/* Get both characters. */
		a = 0;
		b = 0;
		if (sjme_error_is(error = sjme_charSeq_charAt(inSeq,
			at, &a)))
			return sjme_error_default(error);
		if (sjme_error_is(error = sjme_charSeq_charAt(equalsSeq,
			at, &b)))
			return sjme_error_default(error);
		
		/* Are they not the same? */
		if (a != b)
		{
			*outResult = SJME_JNI_FALSE;
			return SJME_ERROR_NONE;
		}
	}
	
	/* There were no failed matches, so they are the same! */
	*outResult = SJME_JNI_TRUE;
	return SJME_ERROR_NONE;
}

sjme_jboolean sjme_charSeq_equalsCharSeqR(
	sjme_attrInNotNull const sjme_charSeq* inSeq,
	sjme_attrInNotNull const sjme_charSeq* equalsSeq)
{
	sjme_jboolean result;
	
	if (inSeq == NULL || equalsSeq == NULL)
		return SJME_JNI_FALSE;
	
	/* Perform the check. */
	result = SJME_JNI_FALSE;
	if (sjme_error_is(sjme_charSeq_equalsCharSeq(inSeq, &result,
		equalsSeq)))
		return SJME_JNI_FALSE;
	
	/* Return whatever result was given. */
	return result;
}

sjme_errorCode sjme_charSeq_equalsUtf(
	sjme_attrInNotNull const sjme_charSeq* inSeq,
	sjme_attrOutNotNull sjme_jboolean* outResult,
	sjme_attrInNotNull sjme_lpcstr equalsUtf)
{
	sjme_errorCode error;
	sjme_charSeq equalsSeq;
	
	if (inSeq == NULL || outResult == NULL || equalsUtf == NULL)
		return SJME_ERROR_NULL_ARGUMENTS;
	
	/* Setup sequence. */
	memset(&equalsSeq, 0, sizeof(equalsSeq));
	if (sjme_error_is(error = sjme_charSeq_newUtfStatic(
		&equalsSeq, equalsUtf, NULL)))
		return sjme_error_default(error);
	
	/* Forward. */
	return sjme_charSeq_equalsCharSeq(inSeq, outResult, &equalsSeq);
}

sjme_jboolean sjme_charSeq_equalsUtfR(
	sjme_attrInNotNull const sjme_charSeq* inSeq,
	sjme_attrInNotNull sjme_lpcstr equalsUtf)
{
	sjme_jboolean result;
	
	if (inSeq == NULL || equalsUtf == NULL)
		return SJME_JNI_FALSE;
	
	/* Perform the check. */
	result = SJME_JNI_FALSE;
	if (sjme_error_is(sjme_charSeq_equalsUtf(inSeq, &result,
		equalsUtf)))
		return SJME_JNI_FALSE;
	
	/* Return whatever result was given. */
	return result;
}

sjme_errorCode sjme_charSeq_hash(
	sjme_attrInNotNull const sjme_charSeq* inSeq,
	sjme_attrOutNotNull sjme_jint* outHash)
{
	sjme_errorCode error;
	sjme_jint result, i, n;
	sjme_jchar c;
	
	if (inSeq == NULL || outHash == NULL)
		return SJME_ERROR_NULL_ARGUMENTS;

	/* How long is the string? */
	n = -1;
	if (sjme_error_is(error = sjme_charSeq_length(inSeq, &n)) || n < 0)
		return sjme_error_default(error);

	/* Calculate for the string. */
	result = 0;
	for (i = 0; i < n; i++)
	{
		/* Get next character. */
		c = 0;
		if (sjme_error_is(error = sjme_charSeq_charAt(inSeq, i, &c)))
			return sjme_error_default(error);
			
		/* Calculate the hashCode(), the JavaDoc gives the following formula: */
		/* == s[0]*31^(n-1) + s[1]*31^(n-2) + ... + s[n-1] .... yikes! */
		result = ((result << 5) - result) + (sjme_jint)c;
	}

	/* Success! */
	*outHash = result;
	return SJME_ERROR_NONE;
}

sjme_errorCode sjme_charSeq_length(
	sjme_attrInNotNull const sjme_charSeq* inSeq,
	sjme_attrOutNotNull sjme_jint* outLen)
{
	if (inSeq == NULL || outLen == NULL)
		return SJME_ERROR_NULL_ARGUMENTS;
	
	if (inSeq->impl == NULL || inSeq->impl->length == NULL)
		return sjme_error_notImplemented(0);
	
	/* Forward. */
	return inSeq->impl->length(inSeq, outLen);
}
	
sjme_errorCode sjme_charSeq_newStatic(
	sjme_attrInNotNull sjme_charSeq* inOutSeq,
	sjme_attrInNotNull const sjme_charSeq_functions* inFunctions,
	sjme_attrInNullable sjme_pointer inOptContext,
	sjme_attrInNullable sjme_frontEnd* inOptFrontEnd)
{
	if (inOutSeq == NULL || inFunctions == NULL)
		return SJME_ERROR_NULL_ARGUMENTS;
	
	/* Clear state. */
	memset(inOutSeq, 0, sizeof(*inOutSeq));
	
	/* Fill in. */
	inOutSeq->impl = inFunctions;
	inOutSeq->context = inOptContext;
	
	/* Copy front end data? */
	if (inOptFrontEnd != NULL)
		memmove(&inOutSeq->frontEnd, inOptFrontEnd,
			sizeof(*inOptFrontEnd));
	
	/* Success! */
	return SJME_ERROR_NONE;
}

sjme_errorCode sjme_charSeq_newUtfStatic(
	sjme_attrInNotNull sjme_charSeq* inOutSeq,
	sjme_attrInNotNull sjme_lpcstr inString,
	sjme_attrInNullable sjme_frontEnd* inOptFrontEnd)
{
	if (inOutSeq == NULL || inString == NULL)
		return SJME_ERROR_NULL_ARGUMENTS;
	
	return sjme_charSeq_newStatic(inOutSeq,
		&sjme_charSeq_basicUtfFunctions,
		(sjme_pointer)inString, inOptFrontEnd);
}

sjme_errorCode sjme_charSeq_startsWithCharSeq(
	sjme_attrInNotNull const sjme_charSeq* inSeq,
	sjme_attrOutNotNull sjme_jboolean* outResult,
	sjme_attrInNotNull const sjme_charSeq* startsWithSeq)
{
	sjme_errorCode error;
	sjme_jint aLen, bLen;
	
	if (inSeq == NULL || outResult == NULL || startsWithSeq == NULL)
		return SJME_ERROR_NULL_ARGUMENTS;
	
	/* Get length of both first. */
	aLen = -1;
	bLen = -1;
	if (sjme_error_is(error = sjme_charSeq_length(inSeq, &aLen)))
		return sjme_error_default(error);
	if (sjme_error_is(error = sjme_charSeq_length(startsWithSeq,
		&bLen)))
		return sjme_error_default(error);
	
	return sjme_error_notImplemented(0);
}

sjme_errorCode sjme_charSeq_startsWithUtf(
	sjme_attrInNotNull const sjme_charSeq* inSeq,
	sjme_attrOutNotNull sjme_jboolean* outResult,
	sjme_attrInNotNull sjme_lpcstr startsWithUtf)
{
	sjme_errorCode error;
	sjme_charSeq startsWithSeq;
	
	if (inSeq == NULL || outResult == NULL || startsWithUtf == NULL)
		return SJME_ERROR_NULL_ARGUMENTS;
	
	/* Setup sequence. */
	memset(&startsWithSeq, 0, sizeof(startsWithSeq));
	if (sjme_error_is(error = sjme_charSeq_newUtfStatic(
		&startsWithSeq, startsWithUtf, NULL)))
		return sjme_error_default(error);
	
	/* Forward. */
	return sjme_charSeq_startsWithCharSeq(inSeq, outResult, &startsWithSeq);
}

sjme_jboolean sjme_charSeq_startsWithUtfR(
	sjme_attrInNotNull const sjme_charSeq* inSeq,
	sjme_attrInNotNull sjme_lpcstr startsWithUtf)
{
	sjme_jboolean result;
	
	if (inSeq == NULL || startsWithUtf == NULL)
		return SJME_JNI_FALSE;
	
	/* Perform the check. */
	result = SJME_JNI_FALSE;
	if (sjme_error_is(sjme_charSeq_startsWithUtf(inSeq, &result,
		startsWithUtf)))
		return SJME_JNI_FALSE;
	
	/* Return whatever result was given. */
	return result;
}
