/* -*- Mode: C; indent-tabs-mode: t; tab-width: 4 -*-
// ---------------------------------------------------------------------------
// SquirrelJME
//     Copyright (C) Stephanie Gawroriski <xer@multiphasicapps.net>
// ---------------------------------------------------------------------------
// SquirrelJME is under the Mozilla Public License Version 2.0.
// See license.mkd for licensing and copyright information.
// -------------------------------------------------------------------------*/

#include <string.h>

#include "lib/scritchui/core/core.h"
#include "lib/scritchui/scritchuiTypes.h"
#include "sjme/debug.h"
#include "sjme/alloc.h"

sjme_errorCode sjme_scritchui_core_containerAdd(
	sjme_attrInNotNull sjme_scritchui inState,
	sjme_attrInNotNull sjme_scritchui_uiComponent inContainer,
	sjme_attrInNotNull sjme_scritchui_uiComponent addComponent)
{
	sjme_scritchui_uiContainer container;
	sjme_list_sjme_scritchui_uiComponent* list;
	sjme_list_sjme_scritchui_uiComponent* newList;
	sjme_errorCode error;
	sjme_jint freeSlot, i, n;
	
	if (inState == NULL || inContainer == NULL || addComponent == NULL)
		return SJME_ERROR_NULL_ARGUMENTS;
	
	/* Cannot add a window to a container despite windows being components. */
	if (addComponent->common.type == SJME_SCRITCHUI_TYPE_WINDOW)
		return SJME_ERROR_INVALID_ARGUMENT;
	
	/* Not implemented? */
	if (inState->impl->containerAdd == NULL)
		return sjme_error_notImplemented(0);
	
	/* Only certain types are containers. */
	if (sjme_error_is(error = inState->intern->getContainer(inState,
		inContainer, &container)) ||
		container == NULL)
		return sjme_error_default(error);
	
	/* Cannot add a component to multiple containers, they must be removed */
	/* manually for consistency purposes. */
	if (addComponent->parent != NULL)
		return SJME_ERROR_ALREADY_IN_CONTAINER;
	
	/* Find free slot in child set. */
	freeSlot = -1;
	list = container->components;
	n = (list == NULL ? 0 : list->length);
	if (list != NULL)
		for (i = 0; i < n; i++)
			if (list->elements[i] == NULL)
			{
				freeSlot = i;
				break;
			}
	
	/* No free slot? Need to alloc list... */
	if (freeSlot < 0)
	{
		/* Setup new list. */
		newList = NULL;
		if (sjme_error_is(error = sjme_list_alloc(inState->pool, n + 4,
			&newList, sjme_scritchui_uiComponent, 0)) || newList == NULL)
			return sjme_error_default(error);
		
		/* Need to copy everything over? */
		if (list != NULL)
			for (i = 0; i < n; i++)
				newList->elements[i] = list->elements[i];
		
		/* Free slot is at the end. */
		freeSlot = n;
		
		/* Set new list. */
		container->components = newList;
		
		/* Free old list. */
		sjme_alloc_free(list);
		
		/* Need to use this one! */
		list = newList;
	}
	
	/* Set free slot in component. */
	list->elements[freeSlot] = addComponent;

	/* Adding to a container sets as visible. */
	addComponent->state.settingVisible = SJME_JNI_TRUE;
	
	/* Forward call. */	
	if (sjme_error_is(error = inState->impl->containerAdd(inState,
		inContainer, container,
		addComponent)))
	{
		/* Undo. */
		list->elements[freeSlot] = NULL;
		
		return sjme_error_default(error);
	}
	
	/* Update parent. */
	addComponent->parent = inContainer;
	
	/* If a component is added to a container which is actually visible */
	/* but the native UI system does not support visibility at this level */
	/* then we need to call the visibility update ourselves. */
	/* That is at the subcomponent granularity and not windows themselves. */
	if (inState->impl->componentSetVisibleListener == NULL)
		if (inContainer->state.isVisible || inContainer->state.isUserVisible)
			inState->intern->updateVisibleComponent(inState,
				addComponent, SJME_JNI_TRUE);
	
	/* Set default bounds for anything that gets added to the container. */
	if (sjme_error_is(error = inState->apiInThread->containerSetBounds(
		inState, inContainer, addComponent,
		0, 0, 8, 8)))
		return sjme_error_default(error);	
	
	/* Revalidate self and the component. */
	if (sjme_error_is(error = inState->apiInThread->componentRevalidate(
		inState, inContainer)))
		return sjme_error_default(error);
	if (sjme_error_is(error = inState->apiInThread->componentRevalidate(
		inState, addComponent)))
		return sjme_error_default(error);
	
	/* If our component has the default focus, then grab it. */
	if (addComponent->common.type == SJME_SCRITCHUI_TYPE_PANEL &&
		((sjme_scritchui_uiPanel)addComponent)->defaultFocus)
		if (sjme_error_is(error = inState->apiInThread->componentFocusGrab(
			inState, addComponent)))
			return sjme_error_default(error);
	
	/* Success! */
	return SJME_ERROR_NONE;
}

sjme_errorCode sjme_scritchui_core_containerRemove(
	sjme_attrInNotNull sjme_scritchui inState,
	sjme_attrInNotNull sjme_scritchui_uiComponent inContainer,
	sjme_attrInNotNull sjme_scritchui_uiComponent removeComponent)
{
	sjme_errorCode error;
	sjme_scritchui_uiContainer container;
	sjme_list_sjme_scritchui_uiComponent* list;
	sjme_jint i, n;
	
	if (inState == NULL || inContainer == NULL || removeComponent == NULL)
		return SJME_ERROR_NULL_ARGUMENTS;
	
	/* Not implemented? */
	if (inState->impl->containerRemove == NULL)
		return sjme_error_notImplemented(0);
		
	/* Only certain types are containers. */
	if (sjme_error_is(error = inState->intern->getContainer(inState,
		inContainer, &container)) ||
		container == NULL)
		return sjme_error_default(error);
	
	/* Get component list, if nothing then it is not here ever. */
	list = container->components;
	if (list == NULL)
		return SJME_ERROR_NOT_IN_CONTAINER;
		
	/* Lose the focus if we are the focused item. */
	if (sjme_error_is(error = inState->intern->bindFocus(inState,
		removeComponent, removeComponent,
		SJME_JNI_FALSE)))
		return sjme_error_default(error);
	
	/* Go through and find the index of the item. */
	for (i = 0, n = list->length; i < n; i++)
	{
		/* Ignore ones that are not this one. */
		if (list->elements[i] != removeComponent)
			continue;

		/* Removing from a container makes it not visible. */
		removeComponent->state.settingVisible = SJME_JNI_FALSE;
		
		/* Forward call. */
		if (sjme_error_is(error = inState->impl->containerRemove(
			inState, inContainer, container, removeComponent)))
			return sjme_error_default(error);
		
		/* Clear component from item set. */
		list->elements[i] = NULL;
		
		/* Remove parent from the component. */
		removeComponent->parent = NULL;
		
		/* Clear set bound. */
		memset(&removeComponent->bounds, 0, 
			sizeof(removeComponent->bounds));
		
		/* Since we removed the component, it is no longer visible. */
		if (inState->impl->componentSetVisibleListener == NULL)
			inState->intern->updateVisibleComponent(inState,
				removeComponent, SJME_JNI_FALSE);
		
		/* Success! */
		return SJME_ERROR_NONE;
	}
	
	/* If this is reached then the component is not here. */
	return SJME_ERROR_NOT_IN_CONTAINER;
}

sjme_errorCode sjme_scritchui_core_containerRemoveAll(
	sjme_attrInNotNull sjme_scritchui inState,
	sjme_attrInNotNull sjme_scritchui_uiComponent inContainer)
{
	sjme_errorCode error;
	sjme_scritchui_uiContainer container;
	sjme_list_sjme_scritchui_uiComponent* list;
	sjme_scritchui_uiComponent remove;
	sjme_jint i, n;
	
	if (inState == NULL || inContainer == NULL)
		return SJME_ERROR_NULL_ARGUMENTS;
	
	/* Not implemented? */
	if (inState->impl->containerRemove == NULL)
		return sjme_error_notImplemented(0);
	
	/* Only certain types are containers. */
	if (sjme_error_is(error = inState->intern->getContainer(inState,
		inContainer, &container)) ||
		container == NULL)
		return sjme_error_default(error);
	
	/* Get component list, if nothing is inside then ignore. */
	list = container->components;
	if (list == NULL)
		return SJME_ERROR_NONE;
	
	/* Go through the list and remove individual components. */
	for (i = 0, n = list->length; i < n; i++)
	{
		/* Skip blank slots. */
		remove = list->elements[i];
		if (remove == NULL)
			continue;
		
		/* Forward removal. */
		if (sjme_error_is(error = inState->apiInThread->containerRemove(
			inState, inContainer, remove)))
			return sjme_error_default(error);
	}
	
	/* Success! */
	return SJME_ERROR_NONE;
}

sjme_errorCode sjme_scritchui_core_containerSetBounds(
	sjme_attrInNotNull sjme_scritchui inState,
	sjme_attrInNotNull sjme_scritchui_uiComponent inContainer,
	sjme_attrInNotNull sjme_scritchui_uiComponent inComponent,
	sjme_attrInPositive sjme_jint x,
	sjme_attrInPositive sjme_jint y,
	sjme_attrInPositiveNonZero sjme_jint width,
	sjme_attrInPositiveNonZero sjme_jint height)
{
	sjme_scritchui_dim suggestDim;
	sjme_errorCode error;
	sjme_scritchui_rect* bound;
	
	if (inState == NULL || inContainer == NULL || inComponent == NULL)
		return SJME_ERROR_NULL_ARGUMENTS;
	
	if (x < 0 || y < 0 || width <= 0 || height <= 0)
		return SJME_ERROR_INVALID_ARGUMENT;
	
	/* This component is not in this specific container. */
	if (inContainer != inComponent->parent)
		return SJME_ERROR_NOT_SUB_COMPONENT;
	
	/* Not implemented? */
	if (inState->impl->containerSetBounds == NULL)
		return sjme_error_notImplemented(0);
	
	/* If the bounds are the same, do nothing. */
	bound = &inComponent->bounds;
	if (x == bound->s.x && y == bound->s.y &&
		width == bound->d.width && height == bound->d.height)
		return SJME_ERROR_NONE; 
	
	/* Store new bounds. */
	bound->s.x = x;
	bound->s.y = y;
	bound->d.width = width;
	bound->d.height = height;
	
	/* Forward call. */
	if (sjme_error_is(error = inState->impl->containerSetBounds(inState,
		inContainer, inComponent, x, y, width, height)))
		return sjme_error_default(error);
	
	/* If we are in a viewport, we can just tell it the dimensions we */
	/* were set to... nothing magical is involved here. */
	memset(&suggestDim, 0, sizeof(suggestDim));
	suggestDim.width = x + width;
	suggestDim.height = y + height;
	if (sjme_error_is(error = inState->intern->viewSuggest(inState,
		inComponent, &suggestDim)))
		return sjme_error_default(error);
	
	/* Success! */
	return SJME_ERROR_NONE;
}

sjme_errorCode sjme_scritchui_core_intern_containerMaxSize(
	sjme_attrInNotNull sjme_scritchui inState,
	sjme_attrInOutNotNull sjme_scritchui_uiComponent inContainer,
	sjme_attrOutNotNull sjme_scritchui_dim* outSize)
{
	sjme_errorCode error;
	sjme_scritchui_uiContainer container;
	sjme_scritchui_uiContainer subContainer;
	sjme_scritchui_dim result, containerSize, viewSize, viewSuggest;
	sjme_scritchui_uiView subView;
	sjme_scritchui_uiComponent item;
	sjme_jint i, n, sw, sh;
	
	if (inState == NULL || inContainer == NULL || outSize == NULL)
		return SJME_ERROR_NULL_ARGUMENTS;

	/* Recover container. */
	container = NULL;
	if (sjme_error_is(error = sjme_scritchui_core_intern_getContainer(
		inState, inContainer,
		&container)) || container == NULL)
		return sjme_error_default(error);

	/* Nothing in this? Ignore. */
	if (container->components == NULL || container->components->length == 0)
	{
		memset(outSize, 0, sizeof(*outSize));
		return SJME_ERROR_NONE;
	}
	
	/* Calculate maximum dimensional size. */
	memset(&result, 0, sizeof(result));
	for (i = 0, n = container->components->length; i < n; i++)
	{
		/* Ignore missing components. */
		item = container->components->elements[i];
		if (item == NULL)
			continue;

		/* Clear base. */
		sw = 0;
		sh = 0;
		
		/* Is this a viewport? */
		subView = NULL;
		memset(&viewSize, 0, sizeof(viewSize));
		memset(&viewSuggest, 0, sizeof(viewSuggest));
		if ((error = inState->intern->getView(
			inState, item, &subView)))
		{
			/* This is a view! */
			if (error == SJME_ERROR_NONE)
			{
				/* Get the current size. */
				viewSize.width = subView->view.d.width;
				viewSize.height = subView->view.d.height;
				
				/* And also the suggested size. */
				viewSuggest.width = subView->lastSuggest.width;
				viewSuggest.height = subView->lastSuggest.height;
			}

			/* Something else? */
			else if (error != SJME_ERROR_INVALID_ARGUMENT)
				return sjme_error_default(error);
		}

		/* Is this another container? */
		subContainer = NULL;
		memset(&containerSize, 0, sizeof(containerSize));
		if ((error = inState->intern->getContainer(
			inState, item, &subContainer)))
		{
			/* This is a container. */
			if (error == SJME_ERROR_NONE)
			{
				/* Recursively get size. */
				if (sjme_error_is(error =
					inState->intern->containerMaxSize(
						inState, item, &containerSize)))
					return sjme_error_default(error);
			}

			/* Something else? */
			else if (error != SJME_ERROR_INVALID_ARGUMENT)
				return sjme_error_default(error);
		}
		
		/* This is a view and either size is valid. */
		if ((viewSuggest.width > 0 && viewSuggest.height > 0) ||
			(viewSize.width > 0 && viewSize.height > 0))
		{
			/* Suggested view size is valid? */
			if (viewSuggest.width > 0 && viewSuggest.height > 0)
			{
				if (viewSuggest.width > sw)
					sw = viewSuggest.width;
				if (viewSuggest.height > sh)
					sh = viewSuggest.height;
			}
	
			/* Standard view size is valid? */
			else
			{
				if (viewSize.width > sw)
					sw = viewSize.width;
				if (viewSize.height > sh)
					sh = viewSize.height;
			}	
		}

		/* Container size is valid? */
		else if (containerSize.width > 0 && containerSize.height > 0)
		{
			/* Add in base size. */
			if (containerSize.width > sw)
				sw = containerSize.width;
			if (containerSize.height > sh)
				sh = containerSize.height;
		}

		/* Item bounds MAY override the view size or the container size! */
		if (item->bounds.d.width > 0 && item->bounds.d.height > 0)
		{
			/* Add in base size. */
			if (item->bounds.d.width > sw)
				sw = item->bounds.d.width;
			if (item->bounds.d.height > sh)
				sh = item->bounds.d.height;
		}

		/* Offset by whatever coordinates the bounds of this has? */
		if (item->bounds.s.x > 0)
			sw += item->bounds.s.x;
		if (item->bounds.s.y > 0)
			sh += item->bounds.s.y;

		/* Greater than the size currently set? */
		if (sw > result.width)
			result.width = sw;
		if (sh > result.height)
			result.height = sh;
	}
	
	/* Success! */
	memmove(outSize, &result, sizeof(*outSize));
	return SJME_ERROR_NONE;
}

sjme_errorCode sjme_scritchui_core_intern_getContainer(
	sjme_attrInNotNull sjme_scritchui inState,
	sjme_attrInNotNull sjme_scritchui_uiComponent inComponent,
	sjme_attrInOutNotNull sjme_scritchui_uiContainer* outContainer)
{
	sjme_scritchui_uiContainer container;
	
	if (inState == NULL || inComponent == NULL || outContainer == NULL)
		return SJME_ERROR_NULL_ARGUMENTS;
	
	/* Only certain types can be a container. */
	switch (inComponent->common.type)
	{
		case SJME_SCRITCHUI_TYPE_PANEL:
			container = &((sjme_scritchui_uiPanel)inComponent)->container;
			break;
			
		case SJME_SCRITCHUI_TYPE_SCROLL_PANEL:
			container = &((sjme_scritchui_uiScrollPanel)
				inComponent)->container;
			break;
		
		case SJME_SCRITCHUI_TYPE_WINDOW:
			container = &((sjme_scritchui_uiWindow)inComponent)->container;
			break;
		
		default:
			return SJME_ERROR_INVALID_ARGUMENT;
	}
	
	/* Success! */
	*outContainer = container;
	return SJME_ERROR_NONE;
}

sjme_errorCode sjme_scritchui_core_intern_updateVisibleContainer(
	sjme_attrInNotNull sjme_scritchui inState,
	sjme_attrInNotNull sjme_scritchui_uiComponent inContainer,
	sjme_attrInValue sjme_jboolean isVisible)
{
	sjme_errorCode error;
	sjme_list_sjme_scritchui_uiComponent* list;
	sjme_scritchui_uiComponent sub;
	sjme_scritchui_uiContainer container;
	sjme_jint i, n;
	
	if (inState == NULL || inContainer == NULL)
		return SJME_ERROR_NULL_ARGUMENTS;
	
	/* We need the container itself. */
	container = NULL;
	if (sjme_error_is(error = inState->intern->getContainer(inState,
		inContainer, &container)) ||
		container == NULL)
		return sjme_error_default(error);
	
	/* Do not fail on the first call! */
	error = SJME_ERROR_NONE;
	
	/* Update our own component. */
	error |= inState->intern->updateVisibleComponent(inState,
		inContainer, isVisible);
	
	/* Need to go through the list. */
	list = container->components;
	if (list != NULL)
	{
		/* Go through everything. */
		for (i = 0, n = list->length; i < n; i++)
		{
			/* If there is no component in this slot, just ignore. */
			sub = list->elements[i];
			if (sub == NULL)
				continue;
			
			/* Send update for callback. */
			error |= inState->intern->updateVisibleComponent(inState,
				sub, isVisible);
		}
		
		/* If the list has changed, due to callback?, free it. */
		if (container->components != list)
		{
			sjme_alloc_free(list);
			list = NULL;
		}
	}
	
	/* Return whatever error state there was. */
	return error;
}
